import logging


class Logging:

    def __init__(self):
        logging.basicConfig(
            level=logging.INFO,
            format="%(asctime)s [%(levelname)s] %(message)s",
            handlers=[
                logging.FileHandler("debug.log"),
                logging.StreamHandler()
            ]
        )

    def debug(self, msg):
        logging.debug(msg)

    def info(self, msg):
        logging.info(msg)



import paramiko
import os


class SftpClientManager(paramiko.SFTPClient):
    """manager for Sftp Client"""

    def put_dir(self, source, target):
        ''' Uploads the contents of the source directory to the target path. The
            target directory needs to exists. All subdirectories in source are
            created under target.
        '''

        for item in os.listdir(source):
            if os.path.isfile(os.path.join(source, item)):
                self.put(os.path.join(source, item), '%s/%s' % (target, item))
            else:
                self.mkdir('%s/%s' % (target, item), ignore_existing=True)
                self.put_dir(os.path.join(source, item), '%s/%s' % (target, item))

    def mkdir(self, path, mode=511, ignore_existing=False):
        ''' Augments mkdir by adding an option to not fail if the folder exists  '''

        try:
            super(SftpClientManager, self).mkdir(path, mode)
        except IOError:
            if ignore_existing:
                pass
            else:
                raise



import json


class Json:

    def __init__(self):
        self.composer_by_path = {}

    def get_variable_from_file(self, var_name, file):
        if file in self.composer_by_path:
            composer = self.composer_by_path[file]
        else:
            with open(file) as f:
                composer = json.load(f)
            self.composer_by_path[file] = composer

        if var_name in composer:
            var_value = composer[var_name]
        else:
            raise Exception('Var name not found.')

        return var_value



from aheadworks_bitbucket_manager.api.bitbucket_api_manager import BitbucketApiManager
from aheadworks_bitbucket_manager.model.data.bitbucket import BitbucketConfig
from aheadworks_core.api.composer_manager import ComposerManager
from aheadworks_core.api.jira_api_manager import JiraApiManager
from aheadworks_core.api.discord_api_manager import DiscordApiManager
from aheadworks_core.api.magento_manager import MagentoManager
from aheadworks_core.api.file_manager import FileManager
from aheadworks_core.model.parser.json import Json as JsonParser
from aheadworks_core.model.http.api_request import ApiRequest
from aheadworks_core.model.data.data_object import DataObject
from datetime import datetime
from urllib.parse import urlparse
import boto3
import json
import os
import re
import shutil
import subprocess

class ReleaseManager:
    """api manager for release"""

    RELEASE_PACK_TASK_LABEL = 'RELEASE-PACK'
    PD_TASK_LABEL = 'PD'
    TEST_TASK_LABEL = 'TEST'

    def __init__(self, jira_api_config):
        self.jira_api_config = jira_api_config
        self.jira_api_manager = JiraApiManager(config=self.jira_api_config)
        self.discord_api_manager = DiscordApiManager()
        self.magento_manager = MagentoManager()
        self.file_manager = FileManager()
        self.json_parser = JsonParser()
        self.aws_s3 = boto3.resource('s3')

    def jira_release(self, jira_project_key, composer_file, discord_bot_url, path_to_files, assign_to):
        module_version = self.json_parser.get_variable_from_file('version', composer_file)
        # module_version = self.magento_manager.get_module_version(path_to_module)

        print(f"jira project key: {jira_project_key}")
        print(f"module version: ' {module_version}")
        print(f"discord bot url: {discord_bot_url}")
        print(f"path to files: {path_to_files}")
        print(f"assign to, account id: {assign_to}")

        if not jira_project_key:
            print('jira_project_key is empty, skip jira release.')
            return False

        jira_instance = self.jira_api_manager.get_jira_instance()

        files_to_upload = []
        file_names = os.listdir(path_to_files)
        for file_name in file_names:
            files_to_upload.append(('file', (file_name, open(f"{path_to_files}/{file_name}", 'rb'))))

        jql = 'labels%3D{}-{}'.format(jira_project_key, module_version)
        links = self.get_release_tasks(jql)

        release_task_key = links[self.RELEASE_PACK_TASK_LABEL]['key']
        pd_task_key = links[self.PD_TASK_LABEL]['key']
        test_task_key = links[self.TEST_TASK_LABEL]['key']

        # add attachments
        self.add_attachments_to_task(release_task_key, files_to_upload)

        # module_dependencies = self.magento_manager.get_module_dependencies(path_to_module)
        module_dependencies = self.magento_manager.get_module_dependencies_from_composer(composer_file)
        composer_package_name = ','.join(list(map(lambda x: x['full_module_name'], module_dependencies.values())))

        # assign release pack to user
        release_issue = jira_instance.issue(release_task_key)
        release_issue.update(assignee={'accountId': assign_to})
        jira_instance.add_comment(release_issue, f'Composer Package Name:\n{composer_package_name}')

        # uncomment if needed check transitions for issue
        # transitions = jira_instance.transitions(release_issue)
        # set done to pd and test issue. transition=31 - Task Done
        pd_issue = jira_instance.issue(pd_task_key)
        jira_instance.transition_issue(pd_issue, transition=31)

        test_issue = jira_instance.issue(test_task_key)
        jira_instance.transition_issue(test_issue, transition=31)

        # release current version
        version = jira_instance.get_project_version_by_name(jira_project_key, module_version)
        release_date = datetime.today().strftime('%Y-%m-%d')
        version.update(released=True, releaseDate=release_date)

        project = jira_instance.project(jira_project_key)

        msg = '{} {}\n'.format(project.name, module_version)
        msg += '\n'.join(list(map(lambda x: x['url'], links.values())))
        msg += '\n' + self.jira_api_manager.get_release_report_all_issues_url(jira_project_key, version.id)

        self.discord_api_manager.send_msg(discord_bot_url, msg)

        return True

    def get_release_tasks(self, jql):
        print(f" https://aheadworks.atlassian.net/issues/?jql={jql}")
        links = dict()
        search_labels = [self.RELEASE_PACK_TASK_LABEL, self.PD_TASK_LABEL, self.TEST_TASK_LABEL]
        tasks = self.jira_api_manager.search_tasks_jql(jql)
        if 'issues' in tasks and len(tasks['issues']):
            for task in tasks['issues']:
                if 'labels' in task['fields']:
                    task_labels = task['fields']['labels']
                    label_intersection = list(set(task_labels) & set(search_labels))
                    if len(label_intersection) == 1:
                        task_key = task['key']
                        task_url = self.jira_api_manager.get_issue_url(task_key)
                        links[label_intersection[0]] = dict({'url': task_url, 'key': task_key})
                    else:
                        raise Exception('Incorrect release labels count.')
                else:
                    raise Exception('Tasks data returned by JIRA has no data on labels')
        else:
            print(tasks)
            raise Exception('Release Tasks not found.')

        return links

    def add_attachments_to_task(self, task_key, files):
        tasks = self.jira_api_manager.add_attachments_to_task(task_key, files)

    def build_swagger_web_api_doc(
            self,
            path_to_module,
            magento_url,
            magento_path_on_server='/var/www/html',
            ssh_port=22,
            ssh_user='root',
            ssh_pass='root'
    ):
        obj = DataObject()
        obj.url = magento_url
        obj.auth_type = 'none'
        magento_request = ApiRequest(obj)

        aws_bucket_name = 'aheadworks_cdn'
        aws_swagger_web_api_doc_path = 'swagger_web_api_doc/'
        magento_app_code_path_on_server = f"{magento_path_on_server}/app/code"
        tmp_dir_m2_modules = '/var/tmp/m2_modules'

        parent_module_name = self.magento_manager.get_module_name(path_to_module)
        if not os.path.isfile(f"{path_to_module}/etc/webapi.xml"):
            return f'Skip Web API doc generation: file etc/webapi.xml has been not found for module {parent_module_name}'

        parsed_url = urlparse(magento_url)

        try:
            list_of_module_paths = self.magento_manager.download_modules_from_git(path_to_module, tmp_dir_m2_modules)
            self.magento_manager.upload_modules_to_server_by_ssh(
                list_of_module_paths['dir_app_code'],
                magento_app_code_path_on_server,
                parsed_url.hostname,
                ssh_port,
                ssh_user,
                ssh_pass
            )
            module_names = ','.join(list_of_module_paths['module_names'])
            magento_request_url = '/generate_web_api_json.php?module_names={}'.format(module_names)
            # workaround, for generate magento cache
            magento_request.get(location=magento_request_url)
            # here we generate json with cache from previous step
            swagger_json = magento_request.get(location=magento_request_url)

            try:
                json.loads(swagger_json)
            except Exception as error:
                print(f"Invalid respone from Swagger:\n{swagger_json}\n\n")
                raise Exception(error)

            s3_result = self.aws_s3.Bucket(aws_bucket_name).put_object(
                Key=f"{aws_swagger_web_api_doc_path}{parent_module_name.lower()}_latest.json",
                Body=swagger_json,
                ACL='public-read'
            )
        except Exception as error:
            raise Exception(error)
        finally:
            os.system(f"rm -rf {tmp_dir_m2_modules}")

        result = f'Web Api Doc Path: https://media.aheadworks.com/{s3_result.key}\n'
        result += f'Magento Request Url: {magento_request_url}\n'
        return result

    def build_ecommerce_pack(self, bitbucket_workspace, bitbucket_repo_slug):
        ComposerManager.init_extra_repos()

        build_dir = os.getcwd()
        artifacts_dir = "/build_archives"
        path = f"{build_dir}/app/code/Aheadworks"
        self.file_manager.create_empty_dir("app/code/Aheadworks")
        self.file_manager.create_empty_dir(artifacts_dir)

        with open('./composer.json') as f:
            composer = json.load(f)
        core_module_name = composer['name']
        core_module_version = composer['version']
        # check if composer have version that is in composer.json
        print("If you see an error just below, then the composer does not see the version of the package.")
        print("Did you tag latest release?")
        subprocess.run(["composer", "show", "-a", core_module_name, core_module_version], check=True)

        module_dependencies = self.magento_manager.get_module_dependencies('./')
        ComposerManager.require_magento_module(core_module_name, composer['version'])

        # @todo use self.magento_manager.download_modules_from_git(path_to_module, tmp_dir_m2_modules)
        magento_module_info = dict()
        artifacts = list()
        # Preinstall suggested modules with composer into /var/www/html/vendor/aheadworks
        for full_module_name, module_item in module_dependencies.items():
            if self.magento_manager.is_suggested_module(build_dir, full_module_name):
                ComposerManager.require_magento_module(full_module_name)

        # full_module_name          vendor/module-name
        # module_name               module-name
        # magento_package_name      ModuleName
        for full_module_name, module_item in module_dependencies.items():
            module_name = module_item['module_name']
            composer_install_path = f"/var/www/html/vendor/aheadworks/{module_name}"
            path_to_composer = f"{composer_install_path}/composer.json"
            path_to_registration = f"{composer_install_path}/registration.php"
            if not os.path.isfile(path_to_composer) or not os.path.isfile(path_to_registration):
                os.system(f"ls {composer_install_path}")
                raise Exception(f"Can not build module {full_module_name}: composer.json or/and registration.php is missing")

            magento_package_name = self._get_magento_package_name(composer_install_path)
            magento_module_info[full_module_name] = magento_package_name
            module_path = f"{path}/{magento_package_name}"

            print(f"Building {full_module_name} as {module_name} in {module_path}...")
            shutil.copytree(composer_install_path, module_path)

            print("\t\u21E8 Cleanup filesystem...")
            self.file_manager.remove_files_and_dirs_ignore_case(
                module_path,
                ['bitbucket-pipelines.yml', 'readme.md', '.gitignore'],
                ['.git']
            )

            print("\t\u21E8 Writing license headers...")
            self.file_manager.add_license_to_php_files(
                module_path,
                magento_package_name,
                self.json_parser.get_variable_from_file('version', f"{module_path}/composer.json")
            )
            os.system(f"echo See https://aheadworks.com/end-user-license-agreement/ >> {module_path}/license.txt")
            artifacts.append(
                self._pack_module(module_path, f"{artifacts_dir}/{module_name}.zip")
            )

        # Now build store packages
        for filename in self._get_store_package_names(magento_module_info[core_module_name], composer['version']):
            artifacts.append(
                self._pack_module(build_dir, f"{artifacts_dir}/{filename}", "app")
            )

        self._upload_artifacts(bitbucket_workspace, bitbucket_repo_slug, artifacts)

    def build_mm_pack(self, bitbucket_workspace, bitbucket_repo_slug):
        # basically marketplace pack is a ecommerce pack where all modules have -subscription postfix in their names
        # so we assume that the ecommerce pack has been built so far
        prebuilt_packages_dir = "/build_archives"
        working_dir = "/tmp/unpack"
        package_dir = "/tmp/packages"
        self.file_manager.create_empty_dir(working_dir)
        self.file_manager.create_empty_dir(package_dir)
        os.system(f"ls -l {prebuilt_packages_dir}")
        # Extract all packages first
        for filename in os.listdir(prebuilt_packages_dir):
            package_fullpath = os.path.join(prebuilt_packages_dir, filename)
            # TODO: process *.zip only
            if os.path.isfile(package_fullpath):
                sources_dir = os.path.join(working_dir, filename)
                print(package_fullpath)
                os.system(f"unzip -q {package_fullpath} -d {sources_dir}")

        # Now gather module names
        module_names = []
        composer_files = self.file_manager.find_all('composer.json', working_dir)
        print(composer_files) # debug
        for composer_file in composer_files:
            package_name = self.json_parser.get_variable_from_file('name', composer_file)
            if not package_name in module_names:
                module_names.append(package_name)

        print(module_names) # debug
        # Replace all found modules names from "vendor/module_name" to "vendor/module_name-subscription"
        # across all composer.json (this includes requires/suggests sections and possibly even more)
        for composer_file in composer_files:
            self._replace_dependencies(composer_file, module_names, "subscription")

        # Create '-subscription' packages
        for originalname in os.listdir(working_dir):
            sources_fullpath = os.path.join(working_dir, originalname)

            split_name, split_extension = os.path.splitext(originalname)
            target_filename = f"{split_name}-subscription{split_extension}"
            target_fullpath = os.path.join(package_dir, target_filename)
            print(target_fullpath) # debug
            self._pack_module(sources_fullpath, target_fullpath)

        # Copy newly built -subscription packages to /build_archives
        os.system(f"cp {package_dir}/* {prebuilt_packages_dir}")

    # Sample metapackage product: https://bitbucket.org/awm2ext/b2b-suite/
    def build_metapackage_pack(self, bitbucket_workspace, bitbucket_repo_slug, artifacts_dir):
        upload_env = f"REPO_URL={os.getenv('AW_COMPOSER_API_URL')} REPO_LOGIN={os.getenv('AW_COMPOSER_API_LOGIN')} REPO_TOKEN={os.getenv('AW_COMPOSER_API_PASSWORD')}"

        # prepare FS
        build_dir = os.getcwd()
        if not artifacts_dir:
            artifacts_dir = "/build_archives"
        path = f"{build_dir}/app/code/Aheadworks"
        self.file_manager.create_empty_dir(artifacts_dir)
        self.file_manager.create_empty_dir(path)

        # Get dependent packages list
        with open("./composer.json") as f:
            composer = json.load(f)

        # Install dependent packages via Composer first
        for full_module_name in composer["require"]:
            ComposerManager.require_magento_module(full_module_name)

        artifacts = []
        for full_module_name in composer["require"]:
            module = full_module_name.split('/')[1]
            metapackage_module_name = f"{module}-{bitbucket_repo_slug}"

            composer_install_path = f"/var/www/html/vendor/aheadworks/{module}"
            magento_package_name = self._get_magento_package_name(composer_install_path)
            module_path = f"{path}/{magento_package_name}"

            print(f"Building {full_module_name} as {module} in {module_path}...")
            shutil.copytree(composer_install_path, module_path)

            print("\t\u21E8 Cleanup filesystem...")
            self.file_manager.remove_files_and_dirs_ignore_case(
                module_path,
                ['bitbucket-pipelines.yml', 'readme.md', '.gitignore'],
                ['.git']
            )

            with open(f"{module_path}/composer.json") as f:
                module_composer = json.load(f)

            print("\t\u21E8 Writing license headers...")
            self.file_manager.add_license_to_php_files(
                module_path,
                magento_package_name,
                module_composer['version']
            )
            os.system(f"echo See https://aheadworks.com/end-user-license-agreement/ >> {module_path}/license.txt")

            old_name = module_composer['name']
            new_name = f"{module_composer['name']}-{bitbucket_repo_slug}"

            print(f"\t\u21E8 Renaming {old_name} to {new_name}...")
            module_composer['name'] = new_name
            with open(f"{module_path}/composer.json", 'w') as f:
                json.dump(module_composer, f, indent=4)

            self._replace_dependencies(f"{module_path}/composer.json", composer["require"], bitbucket_repo_slug)

            zip_package_path = f"{artifacts_dir}/{metapackage_module_name}.zip"
            artifacts.append(
                self._pack_module(module_path, zip_package_path)
            )

            upload_as = f"{metapackage_module_name}-{module_composer['version']}.zip"
            os.system(
                f"echo '{upload_env} python3 -m aheadworks_composer_manager send-package --filename={upload_as} {zip_package_path}' >> {artifacts_dir}/upload")
            # ??? Why we upload to Composer server in this noobish manner instead of using API ???
            composer_server_base_path = f"{os.getenv('AW_COMPOSER_PACKAGES_ROOT')}/aheadworks"
            os.system(
                f"ssh -T {os.getenv('AW_COMPOSER_SSH_URL')} 'mkdir -p {composer_server_base_path}/{metapackage_module_name}'")
            os.system(
                f"scp {zip_package_path} {os.getenv('AW_COMPOSER_SSH_URL')}:{composer_server_base_path}/{metapackage_module_name}/{upload_as}")

        core_version = composer['version']
        with open('composer.json') as f:
            lines = f.readlines()
            result_lines = list()
            for line in lines:
                if line.find('aheadworks/module') != -1:
                    line = line.split(':')[0][:-1] + f'-{bitbucket_repo_slug}"' + f":{line.split(':')[1]}"
                result_lines.append(line)
        with open('composer.json', 'w') as f:
            for line in result_lines:
                f.write(line)

        # build marketplace package
        artifacts.append(
            self._pack_module(".", f"{artifacts_dir}/{bitbucket_repo_slug}.zip", "composer.json")
        )

        # store package
        for filename in self._get_store_package_names(bitbucket_repo_slug, core_version):
            artifacts.append(
                self._pack_module(build_dir, f"{artifacts_dir}/{filename}", "app")
            )

        self._upload_artifacts(bitbucket_workspace, bitbucket_repo_slug, artifacts)

    def _replace_dependencies(self, path_to_composer, module_names, slug):
        fin = open(path_to_composer, "rt")
        data = fin.read()
        fin.close()

        for module_name in module_names:
            # we intentionally quote vendor/package_name string to be "vendor/package_name".
            # otherwise vendor/package_name_subname could be renamed
            # into vendor/package_name-subscription_subname
            # instead of vendor/package_name_subname-subscription
            data = data.replace(f'"{module_name}"', f'"{module_name}-{slug}"')

        fin = open(path_to_composer, "wt")
        fin.write(data)
        fin.close()

    def _pack_module(self, source_module_path, target_package_path, base_dir="."):
        print(f"\t\u21E8 Packing {source_module_path} into {target_package_path}\n")
        base_name, extension = os.path.splitext(target_package_path)
        shutil.make_archive(base_name, extension.lstrip('.'), source_module_path, base_dir)
        return target_package_path

    def _get_store_package_names(self, magento_package_name, version):
        return [
            f"aw_m2_{magento_package_name}-{version}.community_edition.zip",
            f"aw_m2_{magento_package_name}-{version}.enterprise_edition.zip"
        ]

    def _get_magento_package_name(self, module_path):
        with open(f"{module_path}/registration.php") as reg:
            l = reg.readlines()
            for line in l:
                if line.find("Aheadworks_") != -1:
                    m = re.search("Aheadworks_([^\"']+)", line)
                    return m.group(1)
        return ""

    def _upload_artifacts(self, bitbucket_workspace, bitbucket_repo_slug, filepath):
        print(f"\nUploading artifacts\n{filepath}...\n")
        bitbucket_manager = BitbucketApiManager(
            BitbucketConfig(
                bitbucket_workspace=bitbucket_workspace,
                bitbucket_repo_slug=bitbucket_repo_slug
            )
        )
        print(bitbucket_manager.upload_artifacts(filepath))



import requests
import os
import json


class BitbucketConfig:
    def __init__(
            self,
            bitbucket_key=None,
            bitbucket_secret=None,
            bitbucket_workspace=None,
            bitbucket_repo_slug=None,
            auth_header=None
    ):
        self.url = 'https://api.bitbucket.org'
        self.auth_url: str = 'https://bitbucket.org'

        self.bitbucket_key: str = bitbucket_key or os.getenv('BITBUCKET_KEY')
        self.bitbucket_secret: str = bitbucket_secret or os.getenv('BITBUCKET_SECRET')
        self.bitbucket_workspace: str = bitbucket_workspace or os.getenv('BITBUCKET_WORKSPACE')
        self.bitbucket_repo_slug: str = bitbucket_repo_slug or os.getenv('BITBUCKET_REPO_SLUG')
        self.token = self.__get_bitbucket_token(self.bitbucket_key, self.bitbucket_secret)
        self.auth_header = auth_header or ''

    def __get_bitbucket_token(self, key: str, secret: str) -> object:
        """get bitbucket oauth token

        :param key: bitbucket key
        :param secret: bitbucket secret
        :return: token: bitbucket oauth token
        """
        token = ''
        if key and secret:
            data = {'grant_type': 'client_credentials'}
            response = requests.post(
                f'{self.auth_url}/site/oauth2/access_token',
                data=data,
                auth=(key, secret)
            )

            if response.status_code == 200:
                token = json.loads(response.text)['access_token']

        return token



import os


class FileManager:
    """Manager for working with filesystem"""

    def __init__(self):
        pass

    @staticmethod
    def add_info_to_file_header(directories: list, file_extensions: list, info_string: str):
        # todo fix for <?php
        # todo fix for comments for different file types
        log = list()
        for directory in directories:
            for root, dirs, files in os.walk(directory):
                for file in files:
                    if file.split('.')[-1] in file_extensions:
                        with open(os.path.join(root, file)) as f:
                            original_lines = f.readlines()
                        result_lines = list()
                        if original_lines[0].find(info_string) == -1:
                            result_lines.extend(['/*', info_string, '*/', '\n'])
                        for line in original_lines:
                            result_lines.append(line)
                        result = open(os.path.join(root, file), 'w')
                        result.writelines([item for item in result_lines])
                        result.close()
                        log.append(file)
        return log



from setuptools import setup

setup(
    name='aheadworks_deploy_manager',
    version='1.4.29',
    zip_safe=True,
    packages=[
        'aheadworks_composer_manager',
        'aheadworks_core',
        'aheadworks_bitbucket_manager',
        'aheadworks_release_manager',
        'aheadworks_test_manager'
    ],
    package_dir={
        'aheadworks_core': 'aheadworks_core',
        'aheadworks_bitbucket_manager': 'aheadworks_bitbucket_manager',
        'aheadworks_composer_manager': 'aheadworks_composer_manager',
        'aheadworks_release_manager': 'aheadworks_release_manager',
        'aheadworks_test_manager': 'aheadworks_test_manager',
    },
    package_data={
        'aheadworks_core': ['*', '*/*', '*/*/*', '*/*/*/*'],
        'aheadworks_bitbucket_manager': ['*', '*/*', '*/*/*', '*/*/*/*'],
        'aheadworks_composer_manager': ['*', '*/*', '*/*/*', '*/*/*/*'],
        'aheadworks_release_manager': ['*', '*/*', '*/*/*', '*/*/*/*'],
        'aheadworks_test_manager': ['*', '*/*', '*/*/*', '*/*/*/*'],
    },
    include_package_data=True,
    url='https://bitbucket.org/awimpl/deploy-tools',
    license='EULA',
    author='nnuser',
    author_email='sednev@aheadworks.com',
    description='deploy manager from aheadworks',
    install_requires=[
        'fire',
        'python-dateutil',
        'requests',
        'docker',
        'regex',
        'discord>=2.0.0',
        'jira',
        'boto3',
        'paramiko',
        'stringcase',
        'python-dotenv'
    ]
)



#!/usr/bin/env python3
from aheadworks_core.api.composer_manager import ComposerManager
from aheadworks_core.model.cd import Cd as cd
import subprocess
import click
from os.path import join as _
import os
import re
import json
import pathlib
import time
import shutil

def remove_listeners(path):
    with open(path) as phpunit_config:
        data = phpunit_config.read()
        reg = re.compile("<listeners.*</listeners>", re.S)
        data = re.sub(reg, '', data)
    with open(path, 'w') as phpunit_config:
        phpunit_config.write(data)

def di_compile():
    proc = subprocess.run([_(BASIC_PATH, 'bin/magento'), 'module:enable', '--all'])
    ec3 = proc.returncode
    proc = subprocess.run([_(BASIC_PATH, 'bin/magento'), 'setup:di:compile'])
    ec4 = proc.returncode

    print(ec3)
    print(ec4)

    if ec3 or ec4:
        print("Failed to di:compile")

def install(path):
    """
    Install extension(s) to path from path or zip
    """
    click.echo("Installing from %s" % path)

    with open(_(path, 'composer.json')) as f:
        composer = json.load(f)
        repo_name = re.sub(r'[^a-z0-9_]', '_', composer['name'])

    with cd(BASIC_PATH):
        subprocess.run(['composer', 'config', 'repositories.aheadworks', 'composer', 'https://composer.do.staging-box.net'])
        subprocess.run(['composer', 'config', 'http-basic.repo.magento.com', 'bf08744e4b4b3aee1d54fcd7cd56194a', 'f5b232eab5158a4597ecb00f8cacdf4a'])
        subprocess.run(['composer', 'config', 'repositories.' + repo_name, 'path', path])

        proc = subprocess.Popen(['php', '-d', 'memory_limit=4G', '/usr/local/bin/composer', 'require', '--prefer-dist', '{e[name]}:{e[version]}'.format(e=composer)])
        proc.communicate()

        if proc.returncode:
            raise click.ClickException("Failed to install extension")

    result_path = BASIC_PATH / 'vendor' / composer['name']
    return result_path


@click.group()
def cli():
    click.echo("Removing phpunit listeners")
    remove_listeners(BASIC_PATH / 'dev' / 'tests' / 'unit' / 'phpunit.xml.dist')


@cli.command()
@click.option('--severity', default=9, help='Severity level.')
@click.option('--report', default="junit", help='Report type.', type=click.Choice(["full", "xml", "checkstyle", "csv",
                                                                             "json", "junit", "emacs", "source",
                                                                             "summary", "diff", "svnblame", "gitblame",
                                                                             "hgblame", "notifysend"]))
@click.argument('path', type=click.Path(exists=True))
@click.argument('report_file', type=click.Path(), required=False)
def codesniffer(severity, report, path, report_file):
    """Run codesniffer tests against the path"""

    os.system("mkdir -p test-results")
    if os.getenv('SEVERITY'):
        severity = os.getenv('SEVERITY')

    options = [
        _(BASIC_PATH, 'vendor/bin/phpcs'),
        path
    ]

    options += ['--severity=%s' % severity]
    options += ['--standard=Magento2']
    options += ['--extensions=php,phtml']

    stdout = None
    if report_file:
        options += ['--report=' + report]
        stdout = open(report_file, 'w')

    process = subprocess.run(options, stdout=stdout)
    exit(process.returncode)


@cli.command()
@click.option('--report', default="junit", help='Report type.', type=click.Choice(["junit"]))
@click.argument('path', type=click.Path(exists=True))
@click.argument('report_file', type=click.Path(), required=False)
def unit(report, path, report_file):
    """Run unit tests for extension at path"""

    os.system("mkdir -p test-results")
    install(path)
    di_compile()
    
    try:
        os.mkdir('allure')
        shutil.copyfile("/var/www/html/dev/tests/unit/allure/allure.config.php", "allure/allure.config.php")
    except:
        print('allure config not found, its unit test for 2.4.4')
    
    options = [
        _(BASIC_PATH, 'vendor/bin/phpunit'),
        '--configuration',  _(BASIC_PATH, 'dev/tests/unit/phpunit.xml.dist')
    ]

    if report_file:
        options += ['--log-%s' % report, report_file]

    proc = subprocess.Popen(options + [_(path, 'Test/Unit')])
    proc.communicate()

    if not report_file:
        exit(proc.returncode)

    exit(proc.returncode)


@cli.command()
@click.option('--severity', default=0, help='Severity level.')
@click.option('--report', default="junit", help='Report type.', type=click.Choice(["junit"]))
@click.argument('path', type=click.Path(exists=True))
@click.argument('report_file', type=click.Path(), required=False)
def phpstan(severity, report, path, report_file):
    """Run phpstan static analysis against the path"""

    os.system("mkdir -p test-results")
    install(path)
    di_compile()

    options = [
        _(BASIC_PATH, 'vendor/bin/phpstan'),
        'analyse',
        path
    ]

    config = _(path, 'phpstan.neon')
    if pathlib.Path(config).is_file():
        options += ['--configuration', config]

    if os.getenv('PHPSTANSEVERITY'):
        options += ['--level', os.getenv('PHPSTANSEVERITY')]
    else:
        options += ['--level', str(severity)]

    options += [
        '--autoload-file',
        _(BASIC_PATH, 'dev/tests/integration/framework/autoload.php')
    ]

    stdout = None
    if report_file:
        options += ['--error-format', report]
        stdout = open(report_file, 'w')

    process = subprocess.run(options, stdout=stdout)
    exit(process.returncode)


@cli.command()
@click.option('--report', default='xml', help='Report type.', type=click.Choice(['xml']))
@click.argument('path', type=click.Path(exists=True))
@click.argument('report_file', type=click.Path(), required=False)
def mess_detector(report, path, report_file):
    """Run mess detector against the path"""

    os.system("mkdir -p test-results")
    install(path)

    if not report_file:
        report = 'ansi'

    options = [
        _(BASIC_PATH, 'vendor/bin/phpmd'),
        path,
        report,
        _(BASIC_PATH, 'dev/tests/static/testsuite/Magento/Test/Php/_files/phpmd/ruleset.xml')
    ]

    stdout = None
    if report_file:
        stdout = open(report_file, 'w')

    process = subprocess.run(options, stdout=stdout)
    exit(process.returncode)


@cli.command()
@click.option('--report', default="junit", help='Report type.', type=click.Choice(["junit"]))
@click.argument('path', type=click.Path(exists=True))
@click.argument('report_file', type=click.Path(), required=False)
def install_magento(report, path, report_file):
    """Run install magento with extension at path"""

    os.system("mkdir -p test-results")
    install(path)
    global output

    try:
        output = subprocess.run(
            ['sh', "/tmp/create-install-dump.sh", '--withoutdump'],
            check=True,
            text=True
        )
    except subprocess.CalledProcessError as e:
        os.system("mkdir logs1")
        os.system("mysqldump -uroot -proot magento >> logs1/dump.sql")
        os.system(f"cp /var/www/html/var/log/* logs1")
        os.system("cp /var/www/html/app/etc/* logs1")
        exit(e.returncode)

    exit(output.returncode)

@cli.command()
@click.argument('path', type=click.Path(exists=True))
def validate_m2_package(path):
    """
    Test marketplace package
    :param path:
    :return:
    """
    #proc = subprocess.Popen(['php', '-f', '/usr/local/bin/validate_m2_package.php', path])
    #proc.communicate()
    #exit(proc.returncode)


if __name__ == '__main__':
    BASIC_PATH = pathlib.Path(os.environ.get('MAGENTO_ROOT', '/var/www/html'))
    ComposerManager.init_extra_repos()
    cli()



from aheadworks_bitbucket_manager.model.ssh_manager import SshManager
from aheadworks_bitbucket_manager.model.parser.php import Php as PhpParser
from aheadworks_core.model.parser.json import Json as JsonParser
from dotenv import dotenv_values, set_key
import re
import ast
import os
import shutil

class VersionControlManager:

    def __init__(self):
        self.run_app_paths = {}
        self.ssh_manager = SshManager()
        self.php_parser = PhpParser()
        self.json_parser = JsonParser()

    def run_common_app(self, check_containers_arr, up_cmd, host, user):
        command = 'docker ps -q'
        for container_name in check_containers_arr:
            command += ' -f name=' + container_name

        container_ids_arr = self.ssh_manager.run_ssh_command(command, host, user)

        if len(container_ids_arr) != len(check_containers_arr):
            run_result = self.ssh_manager.run_ssh_command(up_cmd, host, user)
            run_result = type(run_result) == list and '\n'.join(run_result)
        else:
            run_result = 'All Common App containers exists'

        return 'run_result:\n' + run_result

    def run_app(
            self,
            path_to_versions,
            path_to_project,
            up_cmd,
            down_exclude_current,
            down_cmd,
            host,
            user
    ):
        """
        :param path_to_versions: path to versions.
        :param path_to_project: path to project folder.
        :param up_cmd: command for run active apps.
        :param down_exclude_current: yes/no, down all app in project.
        :param down_cmd: command for down old apps.
        :param host: host url.
        :param user: host user.
        """
        print('path to versions: ' + path_to_versions)
        print('path to project with releases on server: ' + path_to_project)
        print('run app command: ' + up_cmd)
        print('down all released app in project folder exclude current: ' + down_exclude_current)
        print('kill app command: ' + down_cmd)

        version_config_item_name = ['VERSION', 'SAVE_VERSION']

        save_versions = []
        for var_name in version_config_item_name:
            var_value = self.php_parser.get_variable_from_file(var_name, path_to_versions).strip()
            if var_value:
                save_versions.append(var_value)

        print(f"save versions: {save_versions}")
        dir_items = self.ssh_manager.run_ssh_command('ls ' + path_to_project, host, user)
        print(f"unsorted dir items: {dir_items}")
        # sort dirs by path by desc
        dir_items = self._sort_versions(dir_items)
        print(f"dir items: {dir_items}")

        kill_app_paths = []
        for item in dir_items:
            item = item.strip()
            item_arr = item.split('-')
            item_arr = len(item_arr) == 1 and item_arr.append('') or item_arr
            item_version, item_build = item_arr[:2]
            print("item version: {item_version}")
            if item_version in save_versions and item_version not in self.run_app_paths:
                self.run_app_paths[item_version] = item
            else:
                kill_app_paths.append(item)

        run_commands = []
        for key in self.run_app_paths:
            path_to_project_item = path_to_project + '/' + self.run_app_paths[key]
            run_commands.append(
                'cd ' + path_to_project_item
                + ' && ' + up_cmd
            )
            
        # down all app in project
        kill_result = ''
        down_command = ''
        if down_exclude_current == 'yes':
            kill_commands = []
            for item in kill_app_paths:
                path_to_project_item = path_to_project + '/' + item
                kill_commands.append(
                    'cd ' + path_to_project_item + '; '
                    + down_cmd + '; '
                    + 'rm -rf ' + path_to_project_item
                )

            if kill_commands:
                down_command = ' && '.join(kill_commands)
                kill_result = self.ssh_manager.run_ssh_command(down_command, host, user)
                kill_result = type(kill_result) == list and '\n'.join(kill_result)

        # up all app in project
        run_result = ''
        run_command = ''
        if run_commands:
            run_command = ' && '.join(run_commands)
            run_result = self.ssh_manager.run_ssh_command(run_command, host, user)
            run_result = type(run_result) == list and '\n'.join(run_result)
            
        return '\nrun commands:\n' + str(run_command) \
               + '\n\nkill commands:\n' + str(down_command) \
               + '\n\nrun_result:\n' + str(run_result) \
               + '\n\nkill_result:\n' + str(kill_result)

    def run_app_from_json_config(
            self,
            path_to_versions,
            path_to_project,
            path_to_json_config,
            down_exclude_current,
            host,
            user
    ):
        config_run_commands = self.json_parser.get_variable_from_file('run_commands', path_to_json_config)
        stop_commands = self.json_parser.get_variable_from_file('stop_commands', path_to_json_config)

        run_commands = list()
        check_result = dict({'file': {'commands': list(), 'result': True}})
        for row in config_run_commands:
            command = row['command']
            if 'result' in row:
                result_to = row['result']['to']
                if result_to == 'file':
                    file_name = row['result'][result_to]['name']
                    command = '{} &> {}'.format(command, file_name)
                    check_result['file']['commands'].append(row)
            run_commands.append(command)

        run_commands = ';'.join(run_commands)
        stop_commands = ';'.join(stop_commands)
        run_app_result = self.run_app(
            path_to_versions,
            path_to_project,
            run_commands,
            down_exclude_current,
            stop_commands,
            host,
            user
        )

        check_result_commands = check_result['file']['commands']
        for i, config in enumerate(check_result_commands):
            file = config['result']['file']['name']
            path_to_file = path_to_project + '/' + list(self.run_app_paths.values())[0] + '/' + file
            text_from_file = self.ssh_manager.run_ssh_command('cat ' + path_to_file, host, user)
            index = len(text_from_file) - 1
            while index > 0:
                line_from_file = text_from_file[index].strip()
                if line_from_file:
                    if 'type' in config and config['type'] == 'mftf':
                        if 'OK' in line_from_file:
                            check_result_commands[i]['summary'] = dict({'status': True})
                        else:
                            print(text_from_file)
                            check_result_commands[i]['summary'] = dict({'status': False})
                            check_result['file']['result'] = False
                    break
                index = index - 1

        for result_type in check_result:
            if check_result[result_type]['result'] == False:
                raise Exception('Some commands returned an error!' + '\n\n' + run_app_result)

        return run_app_result

    def get_version_uid(self, version):
        version_uid = re.sub('[.]', '', version)

        return version_uid

    def get_traefik_version_query(self, path_to_versions):
        magento_version_control = self.php_parser.get_variable_from_file('MAGENTO_VERSION_CONTROL', path_to_versions)
        magento_version_control = magento_version_control \
            .replace('[', '{') \
            .replace(']', '}') \
            .replace('=>', ':') \
            .replace('\n', '')
        magento_version_control = ast.literal_eval(magento_version_control)

        app_curr_full_version = self.php_parser.get_variable_from_file('VERSION', path_to_versions)
        app_curr_major_version = app_curr_full_version.split('.')[0:-1]
        app_curr_major_version.append('*')
        app_curr_major_version = '.'.join(app_curr_major_version)
        app_version_map = [app_curr_full_version, app_curr_major_version]

        query = ''
        for module_version, app_version in magento_version_control.items():
            app_version = list(app_version)[0]
            if app_version in app_version_map:
                query = '&& Query(`version={version:' + module_version + '}`)' \
                        + ' || Query(`ver={ver:' + app_version + '}`)' \
                        + ' || HeadersRegexp(`referer`, `(.*)version=' + module_version + '`)' \
                        + ' || HeadersRegexp(`referer`, `(.*)ver=' + app_version + '`)'

        return query

    # --priority_type low or --priority_type height
    def get_free_priority(self, label_name, priority_type, host, user):
        busy_priority = self.ssh_manager.run_ssh_command(
            'docker ps --format \'{{.Label "' + label_name + '"}}\'',
            host,
            user
        )
        filtered_busy_priority = []
        for item in busy_priority:
            item = item.strip()
            if item:
                try:
                    filtered_busy_priority.append(int(item))
                except:
                    pass

        new_priority = 10
        if priority_type == 'low':
            while True:
                if new_priority not in filtered_busy_priority:
                    break
                new_priority += 1
        elif priority_type == 'height':
            if filtered_busy_priority:
                max_priority = max(filtered_busy_priority)
                new_priority = max_priority + 10

        return new_priority

    def get_free_port(self, host, user):
        busy_ports = self.ssh_manager.run_ssh_command(
            'docker ps --format \'{{.Label "aw.apps.app_port_prefix"}}\'',
            host,
            user
        )
        filtered_busy_ports = []
        for item in busy_ports:
            item = item.strip()
            if item:
                try:
                    filtered_busy_ports.append(int(item))
                except:
                    pass

        new_port = 10
        while True:
            if new_port not in filtered_busy_ports:
                break
            new_port += 1

        return new_port

    def modify_env(self, path, template):
        os.system('touch ' + path)
        with open(template, 'r') as f:
            original_lines = f.readlines()
        modify_lines = list()
        for line in original_lines:
            if line.find('=') != -1:
                split_line = line.split('=')
                param = split_line[0]
                value = split_line[1]
                pattern = "%(.*?)%"
                match_values = re.findall(pattern, value)
                match_params = re.findall(pattern, param)
                match_values = match_values + match_params
                if len(match_values):
                    for match_value in match_values:
                        match_value_arr = match_value.split(':-')
                        match_value_arr = len(match_value_arr) == 1 and match_value_arr.append('') or match_value_arr
                        match_value, match_value_default = match_value_arr[:2]
                        if match_value in os.environ and os.environ[match_value]:
                            line = re.sub(pattern, os.environ[match_value], line)
                        elif match_value_default:
                            if match_value_default[0] == '$':
                                match_value_default_env = match_value_default[1:]
                                if match_value_default_env in os.environ and os.environ[match_value_default_env]:
                                    line = re.sub(pattern, os.environ[match_value_default_env], line)
                                else:
                                    line = ''
                            else:
                                line = re.sub(pattern, match_value_default, line)
                        else:
                            line = ''
            modify_lines.append(line)
        nf = open(path, 'w')
        nf.writelines(item for item in modify_lines)
        nf.close()
        return

    def generate_env(self, source_file_path, target_file_path, verbose):
        """
        :param source_file_path: path to the sample .env file
        :param target_file_path: path to the .env file to create
        :param verbose: print all modified values
        """
        # let's copy the source config file first to keep the comments
        try:
            shutil.copy2(source_file_path, target_file_path)
        except FileNotFoundError:
            print("The file \"%s\" does not exist" % (source_file_path))
            raise
        else:
            saved_values = dotenv_values(source_file_path)
            overriden_values = {
                **dotenv_values(source_file_path),
                **os.environ,  # override loaded values with environment variables
            }
            # Now save parsed/overriden values into the target config
            for k,v in saved_values.items():
                new_value = overriden_values[k]
                if isinstance(new_value, str) and new_value.find(' ')!=-1:
                    quote_mode = 'always'
                else:
                    quote_mode = 'never'
                set_key(target_file_path, k, new_value, quote_mode)
                if verbose and v != new_value:
                    print("%s has been changed from \"%s\" to \"%s\"" % (k, v, new_value))

    def _sort_versions(self, items):
        versions = []
        build_versions = []
        for item in items:
            if item.find('-') == -1:
                versions.append(item)
            else:
                build_versions.append(item)
        try:
            versions = self._sort_version_string(versions)
        except Exception:
            versions = []

        try:
            build_versions = self._sort_version_string(build_versions)
        except Exception:
            build_versions = []

        return versions + build_versions

    def _sort_version_string(self, items):
        max_len = max([len(i.split('.')) for i in items])
        return sorted(items, key=lambda s: self._version_to_list(s, max_len), reverse=True)

    def _version_to_list(self, version_string, max_len):
        version = version_string.split('.')
        last = version[-1]
        # check if we have a build number
        if last.find('-') > -1:
            (last, build_num) = last.split('-')
        else:
            build_num = None
        version[-1] = last
        # pad versions so they are the same length
        while len(version)<max_len:
            version.append(0)
        # append a build number if any
        if build_num:
            version.append(build_num)
        # and now convert all items to int
        return [int(i) for i in version]
